/*******************************************************************************
 * Example mongoose server
 *
 * mongooseMgr.c:  Manager for moongoose web server.
 ******************************************************************************/
#define MONGOOSEMGR

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include "main.h"
#include "mongooseMgr.h"
#include "mongoose.h"
#include "css.h"

/*========================================================================
 * STATIC FUNCTIONS
 *=======================================================================*/

/**
 * Pull the UUID from a cookie.
 * Returns a pointer to memory holding the value for the specified cookie
 * or NULL if the cookie is not found.
 * The caller must free the memory pointed to by the return value.
 */
static char *
getCookieParam(const char *cookie, char *param)
{
    char *start = NULL;
    char *end = NULL;
    char *value = NULL;
    int length;

    if ( (cookie!=NULL) && ((start=strstr(cookie, param)) != NULL) )
    {
        if ( (end=strstr(start, "; ")) != NULL )
            length = end-start;
        else
            length = strlen(start);
        value = malloc(length+1);
        memset(value, 0, length+1);
        strncpy(value, start, length);
    }
    return value;
}

/*
 * Redirect to the specified page.
 */
static void
redirect(struct mg_connection *conn, char *uri)
{
    mg_printf(conn, "HTTP/1.1 301 Moved Permanently\r\n"
        "Set-Cookie: UUID=;\r\n"
        "Set-Cookie: LOGIN=;\r\n"
        "Location: %s\r\n\r\n", uri);
}

/*******************************************************************************
 * Cookie based authentication (from mongoose's authenticate.c)
 * taken from http://en.wikipedia.org/wiki/HTTP_cookie#Authentication
 *
 * 1. The user inserts his or her username and password in the text fields
 *    of a login page and sends them to the server;
 * 2. The server receives the username and password and checks them; if
 *    correct, it sends back a page confirming that login has been successful
 *    together with a cookie containing a random session ID that coincides with
 *    a session stored in a database. This cookie is usually made valid for
 *    only the current browser session, however it may also be set to expire at
 *    a future date. The random session ID is then provided on future visits
 *    and provides a way for the server to uniquely identify the browser and
 *    confirm that the browser already has an authenticated user.
 * 3. Every time the user requests a page from the server, the browser
 *    automatically sends the cookie back to the server; the server compares
 *    the cookie with the stored ones; if a match is found, the server knows
 *    which user has requested that page.
 ******************************************************************************/
static void
authorize(struct mg_connection *conn,
          const struct mg_request_info *ri, 
          void *data)
{
    const char  *cookie, *domain;
    char        *uuid = NULL;
    char        *uri = NULL;
    char        *uuidValue = NULL;

    cookie = mg_get_header(conn, "Cookie");
    uri = ri->uri;

    /* Some stuff just gets okay'd without additional checks. */
    if ( (strcmp(uri, "/") == 0) || 
         (strncmp(uri, "/images", 7) == 0) 
       ) 
    {
        /* Always authorize accesses to these pages */
        mg_authorize(conn);
    } 
    else if (strncmp(uri, "/logout", 7) == 0)
    {
        uuid = getCookieParam(cookie, "LOGIN=");
        uuidValue = strtok(uuid,"=");
        uuidValue = strtok(NULL,"=");
        if ( (uuidValue != NULL) && (strncmp(uuidValue, "1", 1) == 0) )
            redirect(conn, "/");
    }
    else if (cookie != NULL && strstr(cookie, "UUID=") != NULL) 
    {
        /* Pull the UUID value from the cookie, if any. */
        uuid = getCookieParam(cookie, "UUID=");
        uuidValue = strtok(uuid,"=");
        uuidValue = strtok(NULL,"=");

        /* Validate the cookie. */
        if ( (uuidValue != NULL) && (strncmp(uuidValue, "MGOOSE", 6) == 0) )
            mg_authorize(conn);
        else 
            redirect(conn, "/logout");
    } 
    else 
    {
        /* Not authorized. Redirect to the front page */
        redirect(conn, uri);
    }
}

/*========================================================================
 * PAGE HANDLERS/CALLBACKS
 *=======================================================================*/

/*
 * This callback function is used to show how to handle 404 error
 */
static void
show404(struct mg_connection *conn,
        const struct mg_request_info *request_info,
        void *user_data)
{
    mg_printf(conn, 
        "HTTP/1.1 200 OK\r\n"
        "Refresh: 5; url=/\r\n"
        "Set-Cookie: UUID=MGOOSE;\r\n"
        "Set-Cookie: LOGIN=;\r\n"
        "Content-Type: text/html\r\n\r\n"
        "<html>\r\n"
        "<head>\r\n"
        FAVICON
        "<title>Mongoose Redirect</title>\r\n"
        "<style type=\"text/css\">\r\n"
        "<!--\r\n"
        "%s\r\n"
        "-->\r\n"
        "</style>\r\n"
        "</head>\r\n"
        "<body>\r\n"
        BANNER
        "<div id=\"content\">\r\n"
        "<br /><br />\r\n"
        "<fieldset>\r\n"
        "<legend>Mongoose Redirect</legend>\r\n"
        "<p class=\"errMsg\">Oops. File not found!<br />\r\n"
        "Clear your cache before trying again.</p>\r\n"
        "</fieldset>\r\n"
        "</div>\r\n"
        "</body>\r\n"
        "</html>\r\n", CSS_BODY
        );
}

/**
 * The front page - the list of functions available to the user.
 */
static void
logout(struct mg_connection *conn,
       const struct mg_request_info *ri, 
       void *data)
{
    mg_printf(conn, 
        "HTTP/1.1 401 Unauthorized\r\n"
        "Set-Cookie: LOGIN=1\r\n"
        "WWW-Authenticate: Digest qop=\"auth\", URI=\"/\", "
        "realm=\"%s\", nonce=\"%lu\"\r\n\r\n",
        REALM, (unsigned long) time(NULL));
}

/**
 * The front page - the list of functions available to the user.
 */
static void
pageFront(struct mg_connection *conn,
          const struct mg_request_info *ri, 
          void *data)
{
    const char *cookie;
    char       *uuid      = NULL;
    char       *uuidValue = NULL;

    /* Pull the UUID from the cookie, if any. */
    cookie = mg_get_header(conn, "Cookie");
    if (cookie != NULL && strstr(cookie, "UUID=") != NULL)
    {
        uuid = getCookieParam(cookie, "UUID=");
        uuidValue = strtok(uuid,"=");
        uuidValue = strtok(NULL,"=");
    }
    if ( uuidValue == NULL )
        uuidValue = "MGOOSE";

    /* Send the front page list of commands. */
    mg_printf(conn, "HTTP/1.1 200 OK\r\n"
        "Set-Cookie: UUID=%s;\r\n"
        "Set-Cookie: LOGIN=;\r\n"
        "content-Type: text/html\r\n\r\n"
        "<html>\r\n"
        "<head>\r\n"
        FAVICON
        "<title>Mongoose Server</title>\r\n"
        "<style type=\"text/css\">\r\n"
        "<!--\r\n"
        "%s\r\n"
        "-->\r\n"
        "</style>\r\n"
        "</head>\r\n"
        "<body>\r\n"
        BANNER
        "<div id=\"content\">\r\n"
        "<br /><br />\r\n"
        "<fieldset>\r\n"
        "<legend> Main Menu</legend> \n"
        "<ul> \n"
        "    <li> <a href=\"/hostInfo\">Host Info</a> </li>\n"
        "</ul> \n"
        "</fieldset> \n"
        "</div>\r\n"
        "</body>\n"
        "</html>\n",
        uuidValue,
        CSS_BODY
        );
}

/**
 * List the set of registerd slaves.
 */
static void
hostInfo(struct mg_connection *conn,
              const struct mg_request_info *ri, 
              void *data)
{
    const char *cookie;
    char       *uuid      = NULL;
    char       *uuidValue = NULL;

    char        item[512];
    char        *itemTemplate = "<li>%s</li>";
    struct utsname uname_buf;

    /* Pull the UUID from the cookie, if any. */
    cookie = mg_get_header(conn, "Cookie");
    if (cookie != NULL && strstr(cookie, "UUID=") != NULL)
    {
        uuid = getCookieParam(cookie, "UUID=");
        uuidValue = strtok(uuid,"=");
        uuidValue = strtok(NULL,"=");
    }
    if ( uuidValue == NULL )
        uuidValue = "";

    uname(&uname_buf);
    sprintf(item, itemTemplate, uname_buf.nodename);

    /* Build and send the page. */
    mg_printf(conn, "HTTP/1.1 200 OK\r\n"
        "Set-Cookie: UUID=%s;\r\n"
        "Set-Cookie: LOGIN=;\r\n"
        "content-Type: text/html\r\n\r\n"
        "<html>\n"
        "<head>\r\n"
        FAVICON
        "<title>Host Info</title>\r\n"
        "<style type=\"text/css\">\r\n"
        "<!--\r\n"
        "%s\r\n"
        "-->\r\n"
        "</style>\r\n"
        "</head>\r\n"
        "<body>\r\n"
        BANNER
        "<div id=\"content\">\r\n"
        "<br /><br />\r\n"
        "<fieldset>\r\n"
        "<legend> Host Info</legend> \n"
        "<ul> \n"
        "%s"
        "</ul> \n"
        "<ul class=\"nav\"> \n"
        "    <li> <a href=\"/\">Main Menu</a> </li>\n"
        "    <li> <a href=\"/logout\">Logout</a> </li>\n"
        "</ul> \n"
        "</fieldset> \n"
        "</div>\r\n"
        "</body>\n"
        "</html>\n", 
        uuidValue,
        CSS_BODY,
        item
        );
}

/**
 * Initialize the mongoose web server.
 * This method:
 * 1. Initializes the mongoose context
 * 2. Sets the port and starts the server.
 * 3. Sets up the callbacks for RESTful requests
 */
void 
mongooseMgrInit()
{
    struct mg_context *ctx;
    char htpath[PATH_MAX];

    ctx = mg_start();
    memset(htpath, 0, PATH_MAX);
    strcpy(htpath, documentRoot);
    strcat(htpath, "/.htpasswd");

    /* Set the options for the server. */
    mg_set_option(ctx, "ports", port);
    mg_set_option(ctx, "root", documentRoot);
    mg_set_option(ctx, "dir_list", "0");
    mg_set_option(ctx, "auth_gpass", htpath);
    mg_set_option(ctx, "auth_realm", REALM);
    mg_set_option(ctx, "error_log", "errors.log");
    mg_set_option(ctx, "access_log", "access.log");

    /* Set the callback for user authentication. */
    mg_set_auth_callback(ctx, "/*", &authorize, NULL);

    /* Set the callbacks for specific pages. */
    mg_set_uri_callback(ctx, "/", &pageFront, NULL);
    mg_set_uri_callback(ctx, "/hostInfo", &hostInfo, NULL);
    mg_set_uri_callback(ctx, "/logout", &logout, NULL);

    /* Error handlers */
    mg_set_error_callback(ctx, 404, show404, NULL);
}
